package othello.view;

import java.rmi.* ;
import java.rmi.server.UnicastRemoteObject;
import java.util.Observer ;
import java.util.Observable ;

import othello.model.RemoteGameModelInterface;
import othello.model.Pair;
import othello.model.GameConstants;
import othello.observation.* ;

/** The animator object caches the state of the board
 *  and animates transitions between states of the board.
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class Animator extends UnicastRemoteObject
implements RemoteConcurrentObserver, GameConstants
{
	private static final long serialVersionUID = 8361501772785269040L;

	// If Java supported multiple inheritance then it would
	// extend both java.util.Observable and java.rmi.server.UnicastRemoteObject.
	// Since that is not possible, we delegate to a private Observable object.
	private static class MyObservable extends Observable  {
		void setChangedBackDoor() { super.setChanged() ; }
	}
	
	private MyObservable observableDelegate = new MyObservable() ;

	private int [][] square = new int[ROWS][COLS] ;
	{
		for( int r=0 ; r<ROWS ; ++r ) {
			for( int c=0 ; c<COLS ; ++c ) {
				square[r][c] = EMPTY ; } }
	}

	public final int QUIET_STATE = 0 ;
	public final int FLYING_STATE = 1 ;
	public final int FLIPPING_STATE = 2 ;

	private int state = QUIET_STATE ;
	private int flippingRow ;
	private int flippingCol ;
	private double amountOfFlip ;
	private double flyingRow ;
	private double flyingCol ;
	private int newColour ;

	private RemoteGameModelInterface gameModel ;

	public Animator( RemoteGameModelInterface gameModel )
			throws RemoteException {
		this.gameModel = gameModel ;
		System.err.println("Animator adds itself "+this+" as observer");
		gameModel.addObserver(this);}

	public void addObserver( Observer observer ) {
		observableDelegate.addObserver( observer ); }

	public void update( RemoteCountMonitorInterface countMonitor, Object arg ) {
		Pair pair = (Pair) arg ;
		int row = pair.getRow() ;
		int col = pair.getCol() ;
		int oldColour ; synchronized( Animator.this ) {
			oldColour = square[ row ][ col ] ; }
		int newColour ;
		try {
			newColour = gameModel.getPieceAt( row, col ) ; }
		catch( RemoteException e ) {
			System.err.println( "Lost communication with server" );
			e.printStackTrace( System.err );
			return ; }
		if( oldColour == EMPTY ) {
			// This is a new piece being layed.
			animateFlyIn( row, col, newColour, countMonitor ) ;
		}
		else {
			// This is a flip.
			animateFlip( row, col, newColour, countMonitor ) ;
		}

	}

	synchronized public int getPieceAt( int r, int c ) { return square[r][c] ; }

	synchronized public int getState() { return state ; }

	/** Flipping amount varies from 0 (not at all) to 1 (fully flipped) */
	synchronized public double getFlippingAmount() { return amountOfFlip ; }

	synchronized public int getFlippingRow() { return flippingRow ; }

	synchronized public int getFlippingCol() { return flippingCol ; }

	synchronized public double getFlyingRow() { return flyingRow ; }

	synchronized public double getFlyingCol() { return flyingCol ; }

	synchronized public int getNewColour() { return newColour ; }

	private void animateFlyIn( final int row, final int col,
			final int newColour,
			final RemoteCountMonitorInterface counter ) {
		class FlyThread extends Thread {
			@Override public void run() {
				int iterations ;
				synchronized( Animator.this ) {
					state = FLYING_STATE ;
					Animator.this.newColour = newColour ;
					if( newColour == WHITE ) {
						flyingRow = -2 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+row) ; }
					else {
						flyingRow = ROWS + 1 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+9-row) ; } }

				double dr = (row - flyingRow) / iterations ;
				double dc = (col - flyingCol) / iterations ;

				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						flyingRow += dr ;
						flyingCol += dc ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers();
				try {
					counter.decrement(); }
				catch( RemoteException e ) {
					e.printStackTrace( System.err );
					System.err.println( "Lost communication with server" ); } }
		}
		Thread t = new FlyThread() ;
		t.start();
	}

	private void animateFlip( final int row, final int col,
			final int newColour,
			final RemoteCountMonitorInterface counter ) {
		class FlipThread extends Thread {
			@Override public void run() {
				synchronized( Animator.this ) {
					state = FLIPPING_STATE ;
					Animator.this.newColour = newColour ;
					square[ row ][ col ] = EMPTY ;
					flippingRow = row ;
					flippingCol = col ;
					amountOfFlip = 0.0 ; }

				int iterations = 50 ;
				double delta =  1 / (double)iterations ;
				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						amountOfFlip += delta ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers();
				try {
					counter.decrement(); }
				catch( RemoteException e ) {
					e.printStackTrace( System.err );
					System.err.println( "Lost communication with server" ); } }
		}
		Thread t = new FlipThread() ;
		t.start();
	}

	private void setChanged() { observableDelegate.setChangedBackDoor() ; }

	private void notifyObservers() { observableDelegate.notifyObservers(); }
}
